//JVS-Bridge Public Release: Copyright 2022 Tucson Logic LLC.
//This program may be distributed under the terms of the GNU GPL v3.
//See the LICENSE file for more information.

//jvsbridge-daemon.c
//Unattended program for driving JVS-Bridge from Linux inputs
//Bryan E. Topp <betopp@betopp.com> 2020

#include <stdlib.h>
#include <stdint.h>
#include <poll.h>
#include <string.h>
#include <stdbool.h>
#include <unistd.h>
#include <fcntl.h>
#include <termios.h>
#include <sys/ioctl.h>
#include <errno.h>
#include <error.h>
#include <stdio.h>
#include <assert.h>
#include <linux/joystick.h>

//Manual "Release version" to show
#define RELEASEVERSION "1.0"

//Should have been provided by Makefile
#ifndef BUILDVERSION
	#define BUILDVERSION "Unknown"
#endif
#ifndef BUILDUSER
	#define BUILDUSER "Unknown"
#endif
#ifndef BUILDDATE
	#define BUILDDATE "Unknown"
#endif

//Temporary buffer for making JVS commands
#define COMMANDBUF_LEN 64
uint8_t commandbuf[COMMANDBUF_LEN];

//User switches that JVS supports, per seat.
typedef enum jvs_usrsw_e
{
	JVS_USRSW_PUSH2  =  0,
	JVS_USRSW_PUSH1  =  1,
	JVS_USRSW_RIGHT  =  2,
	JVS_USRSW_LEFT   =  3,
	JVS_USRSW_DOWN   =  4,
	JVS_USRSW_UP     =  5,
	JVS_USRSW_SVC    =  6,
	JVS_USRSW_START  =  7,
	
	JVS_USRSW_PUSH10 =  8,
	JVS_USRSW_PUSH9  =  9,
	JVS_USRSW_PUSH8  = 10,
	JVS_USRSW_PUSH7  = 11,
	JVS_USRSW_PUSH6  = 12,
	JVS_USRSW_PUSH5  = 13,
	JVS_USRSW_PUSH4  = 14,
	JVS_USRSW_PUSH3  = 15,
	
	JVS_USRSW_MAX
	
} jvs_usrsw_t;

//System switches that JVS supports, overall
typedef enum jvs_syssw_e
{
	JVS_SYSSW_RESERVED4 = 0,
	JVS_SYSSW_RESERVED3 = 1,
	JVS_SYSSW_RESERVED2 = 2,
	JVS_SYSSW_RESERVED1 = 3,
	JVS_SYSSW_TILT3     = 4,
	JVS_SYSSW_TILT2     = 5,
	JVS_SYSSW_TILT1     = 6,
	JVS_SYSSW_TEST      = 7,
	
	JVS_SYSSW_MAX
} jvs_syssw_t;

//Names of each user-switch, for reading from the mapping file
const char *jvs_usrsw_names[JVS_USRSW_MAX] = 
{
	[JVS_USRSW_PUSH2] = "push2",
	[JVS_USRSW_PUSH1] = "push1",
	[JVS_USRSW_RIGHT] = "right",
	[JVS_USRSW_LEFT] = "left",
	[JVS_USRSW_DOWN] = "down",
	[JVS_USRSW_UP] = "up",
	[JVS_USRSW_SVC] = "svc",
	[JVS_USRSW_START] = "start",
	[JVS_USRSW_PUSH10] = "push10",
	[JVS_USRSW_PUSH9] = "push9",
	[JVS_USRSW_PUSH8] = "push8",
	[JVS_USRSW_PUSH7] = "push7",
	[JVS_USRSW_PUSH6] = "push6",
	[JVS_USRSW_PUSH5] = "push5",
	[JVS_USRSW_PUSH4] = "push4",
	[JVS_USRSW_PUSH3] = "push3",
};

//Names of each system-switch, for reading from the mapping file
const char *jvs_syssw_names[JVS_SYSSW_MAX] = 
{
	[JVS_SYSSW_RESERVED4] = "reserved4",
	[JVS_SYSSW_RESERVED3] = "reserved3",
	[JVS_SYSSW_RESERVED2] = "reserved2",
	[JVS_SYSSW_RESERVED1] = "reserved1",
	[JVS_SYSSW_TILT3] = "tilt3",
	[JVS_SYSSW_TILT2] = "tilt2",
	[JVS_SYSSW_TILT1] = "tilt1",
	[JVS_SYSSW_TEST] = "test",
};

//Linux joysticks that we know about
#define JOY_MAX 4
#define AXIS_MAX 32
#define BUTTON_MAX 32
typedef struct joy_instance_s
{
	//Whether this is a valid entry in the array
	bool valid;
	
	//Number of joystick known in filesystem (i.e., /dev/js___)
	int num;
	
	//File descriptor of joystick as opened.
	int fd;
	
	//Player that this joystick represents, for sending to JVS
	int player;

	//Last state of axes, in joystick order
	int16_t axis_state[AXIS_MAX];
	
	//Whether axis-as-switch is currently activated in the negative direction
	bool daxis_neg_act[AXIS_MAX];
	
	//Whether axis-as-switch is currently activated in the positive direction
	bool daxis_pos_act[AXIS_MAX];
	
	//Last state of buttons, in joystick order
	bool button_state[BUTTON_MAX];
	
	//Index of the joycfg_t for this joystick device. -1 if not found.
	int joycfg;
	
	//Index of the mapping_t currently used. -1 if not found.
	int mapping;
	
} joy_instance_t;
joy_instance_t joy_instance_array[JOY_MAX];

//Joystick mappings that we have read from the mapping file.
#define MAPPING_MAX 64
typedef struct mapping_s
{
	//JVS user switch mask that each button triggers
	uint32_t usrsw_for_button[BUTTON_MAX];
	
	//JVS system switch mask that each button triggers
	uint32_t syssw_for_button[BUTTON_MAX];
	
	//"Digital axis" mapping - an analog axis on the Linux joystick to a set of JVS switches.
	
	int16_t daxis_neg_act[AXIS_MAX]; //Activation threshold in negative direction
	int16_t daxis_neg_deact[AXIS_MAX]; //Deactivation threshold in negative direction
	uint32_t daxis_neg_usrsw[AXIS_MAX]; //Switches to trigger when the axis is activated negatively
	
	int16_t daxis_pos_act[AXIS_MAX]; //Activation threshold in positive direction
	int16_t daxis_pos_deact[AXIS_MAX]; //Deactivation threshold in positive direction
	uint32_t daxis_pos_usrsw[AXIS_MAX]; //Switches to trigger when the axis is activated positively
	
	//Chord to activate this mapping, if there are multiple mappings for the same joystick device.
	uint32_t activation_buttons_mask;
	
} mapping_t;
mapping_t mapping_array[MAPPING_MAX];
int mapping_count;

//Joystick device names that we have read from the mapping file.
#define JOYCFG_MAX 32
#define JOYCFG_NAME_BUFLEN 128
typedef struct joycfg_s
{
	//Name of joystick to match
	char name[JOYCFG_NAME_BUFLEN];
	
	//Mappings configured for the device
	int mapping_first;
	int mapping_count;
	
} joycfg_t;
joycfg_t joycfg_array[JOYCFG_MAX];
int joycfg_count;


//Reads the given map-file and fills out mapping data
void mapfile_read(const char *filename)
{
	//Zero out existing mappings
	memset(mapping_array, 0, sizeof(mapping_array));
	mapping_count = 0;
	
	memset(joycfg_array, 0, sizeof(joycfg_array));
	joycfg_count = 0;
	
	//Make a dummy, default mapping, which maps everything in-order.
	for(int bb = 0; bb < BUTTON_MAX; bb++)
	{
		if(bb < JVS_USRSW_MAX)
		{
			mapping_array[0].usrsw_for_button[bb] |= (1 << bb);
		}
		else
		{
			mapping_array[0].syssw_for_button[bb] |= (1 << (bb - JVS_USRSW_MAX));
		}
	}
	mapping_count = 1;
	
	//Make a dummy, default device that uses the default mapping
	strcpy(joycfg_array[0].name, "Default");
	joycfg_array[0].mapping_first = 0;
	joycfg_array[0].mapping_count = 1;
	joycfg_count = 1;
	
	//Open the specified mapping file
	FILE *mapfile = fopen(filename, "r");
	if(mapfile == NULL)
	{
		perror("mapfile fopen");
		exit(-1);
	}
	
	//Work through each line in the mapping file.
	while(!feof(mapfile))
	{
		char linebuf[256] = {0};
		char stringbuf[256] = {0};
		char charbuf = '\0';
		int intbufs[4] = { -1, -1, -1, -1 };
		if(fgets(linebuf, sizeof(linebuf) - 1, mapfile) == NULL)
			break;
		
		if(sscanf(linebuf, " js : \"%250[^\"]\" ", stringbuf) == 1)
		{
			//Start of a new device entry.
			assert(joycfg_count < JOYCFG_MAX);
			joycfg_array[joycfg_count].mapping_first = mapping_count;
			joycfg_array[joycfg_count].mapping_count = 0;
			strncpy(joycfg_array[joycfg_count].name, stringbuf, sizeof(joycfg_array[joycfg_count].name) - 1);
			joycfg_count++;
		}
		else if(sscanf(linebuf, " map : %d , %d , %d , %d ", &(intbufs[0]), &(intbufs[1]), &(intbufs[2]), &(intbufs[3])) > 0 )
		{
			//Start of a new mapping within a device entry.
			//The latest device entry is used.
			assert(mapping_count < MAPPING_MAX);
			assert(joycfg_count > 0);
			memset(mapping_array[mapping_count].usrsw_for_button, 0, sizeof(mapping_array[mapping_count].usrsw_for_button));
			memset(mapping_array[mapping_count].syssw_for_button, 0, sizeof(mapping_array[mapping_count].syssw_for_button));
			
			//The numbers specified on this line are buttons used to activate the mapping, if all pressed.
			mapping_array[mapping_count].activation_buttons_mask = 0;
			for(int bb = 0; bb < 4; bb++)
			{
				if(intbufs[bb] >= 0)
				{
					mapping_array[mapping_count].activation_buttons_mask |= 1 << intbufs[bb];
				}
			}
			
			joycfg_array[joycfg_count - 1].mapping_count++;
			mapping_count++;
		}
		else if(sscanf(linebuf, " button : %d = %s ", &(intbufs[0]), stringbuf) == 2)
		{
			//Button mapping in the current mapping.
			int button_to_map = intbufs[0];
			
			//Find what JVS button it refers to
			int usrsw = -1;
			int syssw = -1;
			for(int ss = 0; ss < JVS_USRSW_MAX; ss++)
			{
				if(!strcmp(jvs_usrsw_names[ss], stringbuf))
				{
					usrsw = ss;
					break;
				}
			}
			for(int ss = 0; ss < JVS_SYSSW_MAX; ss++)
			{
				if(!strcmp(jvs_syssw_names[ss], stringbuf))
				{
					syssw = ss;
					break;
				}
			}
			
			//Set the bit in the mappings
			assert(mapping_count > 0);
			assert(button_to_map < BUTTON_MAX);
			
			if(usrsw >= 0)
				mapping_array[mapping_count - 1].usrsw_for_button[button_to_map] |= 1 << usrsw;
			
			if(syssw >= 0)
				mapping_array[mapping_count - 1].syssw_for_button[button_to_map] |= 1 << syssw;
		}
		else if(sscanf(linebuf, " daxis : %d %[<>] %d , %d = %s ", 
			&(intbufs[0]), &charbuf, &(intbufs[1]), &(intbufs[2]), stringbuf) == 5)
		{
			//Digital axis mapping - Linux axis to JVS switches, via thresholds. Negative side.
			int axis_to_map = intbufs[0];
			int activation = intbufs[1];
			int deactivation = intbufs[2];
			bool negside = (charbuf == '<');
			
			//Find what JVS button it refers to
			int usrsw = -1;
			for(int ss = 0; ss < JVS_USRSW_MAX; ss++)
			{
				if(!strcmp(jvs_usrsw_names[ss], stringbuf))
				{
					usrsw = ss;
					break;
				}
			}
		
			assert(axis_to_map >= 0 && axis_to_map < AXIS_MAX);
			assert(mapping_count > 0);
			
			if(negside)
			{
				mapping_array[mapping_count - 1].daxis_neg_act[axis_to_map] = activation;
				mapping_array[mapping_count - 1].daxis_neg_deact[axis_to_map] = deactivation;
				mapping_array[mapping_count - 1].daxis_neg_usrsw[axis_to_map] |= 1 << usrsw;
			}
			else
			{
				mapping_array[mapping_count - 1].daxis_pos_act[axis_to_map] = activation;
				mapping_array[mapping_count - 1].daxis_pos_deact[axis_to_map] = deactivation;
				mapping_array[mapping_count - 1].daxis_pos_usrsw[axis_to_map] |= 1 << usrsw;
			}
		}
	}
}



//Looks for new joysticks to open, tries to open them and add them to the array of known joysticks.
void joy_find(void)
{
	//Find out the highest-numbered device we currently have open, and if we have room for any more.
	int max_opened = 0;
	int free_slot = -1;
	for(int jj = 0; jj < JOY_MAX; jj++)
	{		
		if(joy_instance_array[jj].valid && (joy_instance_array[jj].num > max_opened))
			max_opened = joy_instance_array[jj].num;
		
		if((!joy_instance_array[jj].valid) && (free_slot == -1))
			free_slot = jj;
	}
	
	//If we don't have room for more joysticks, don't try to open any.
	if(free_slot == -1)
		return;
	
	//Find the first joystick we don't have opened, but can now open.
	for(int jsnum = 0; jsnum <= max_opened + 1; jsnum++)
	{
		//If we already have the device opened, don't duplicate it
		bool already_opened = false;
		for(int jj = 0; jj < JOY_MAX; jj++)
		{
			if(joy_instance_array[jj].valid && joy_instance_array[jj].num == jsnum)
			{
				already_opened = true;
				break;
			}
		}
		
		if(already_opened)
			continue;
		
		//Try to open the joystick device, and ignore it if it's not openable.
		char namebuf[256] = {0};
		snprintf(namebuf, sizeof(namebuf) - 1, "/dev/input/js%d", jsnum);
		int newfd = open(namebuf, O_RDONLY);
		if(newfd == -1)
			continue;
		
		//Do an IOCTL to figure out the name of the joystick, and see if we know of a config for it.
		//Use config/mapping 0 by default - these are set up as a fallback.
		int joycfg = 0;
		int mapping = 0;
		if(ioctl(newfd, JSIOCGNAME(sizeof(namebuf)-1), namebuf) >= 0)
		{
			for(int cc = 1; cc < joycfg_count; cc++) //Start at 1 - skip over the default 0 entry
			{
				if(!strcmp(joycfg_array[cc].name, namebuf))
				{
					joycfg = cc;
					mapping = joycfg_array[cc].mapping_first;
					break;
				}
			}
		}
		
		//Figure out what player number it will be
		int newplayer = 0;
		while(1)
		{
			bool player_exists = false;
			
			for(int jj = 0; jj < JOY_MAX; jj++)
			{
				if(joy_instance_array[jj].valid && (joy_instance_array[jj].player == newplayer))
				{
					player_exists = true;
					break;
				}
			}
			
			if(player_exists)
			{
				newplayer++;
			}
			else
			{
				break;
			}
		}
		
		//Put it in the free slot we found earlier.
		joy_instance_array[free_slot].valid = true;
		joy_instance_array[free_slot].num = jsnum;
		joy_instance_array[free_slot].fd = newfd;
		joy_instance_array[free_slot].player = newplayer;
		joy_instance_array[free_slot].joycfg = joycfg;
		joy_instance_array[free_slot].mapping = mapping;
		memset(joy_instance_array[free_slot].axis_state, 0, sizeof(joy_instance_array[free_slot].axis_state));
		memset(joy_instance_array[free_slot].button_state, 0, sizeof(joy_instance_array[free_slot].button_state));
		printf("Opened joystick %d and it will be player %d.\n", jsnum, newplayer);
		if(joycfg >= 0)
			printf("The joystick has a known configuration [%s].\n", joycfg_array[joycfg].name);
		else
			printf("The joystick is not configured and will use a default configuration.\n");
		
		break;
	}
}

//Updates joystick data from known devices
void joy_update(void)
{
	//Use the poll call, to perform a read from any one of the joysticks.
	struct pollfd pollfds_array[JOY_MAX];
	memset(pollfds_array, 0, sizeof(pollfds_array));
	int pollfds_num = 0;
	for(int jj = 0; jj < JOY_MAX; jj++)
	{
		if(joy_instance_array[jj].valid)
		{
			pollfds_array[pollfds_num].fd = joy_instance_array[jj].fd;
			pollfds_array[pollfds_num].events = POLLIN;
			pollfds_num++;
		}
	}
	
	int pollret = poll(pollfds_array, pollfds_num, 10);
	if(pollret == -1)
	{
		//Error happened
		perror("poll");
		exit(-1);
	}
	
	if(pollret == 0)
	{
		//Nothing happened - timeout or signal before any data was ready.
		return;
	}
	
	//Look for the file descriptors that were ready, and read events from them.
	int pollfd_next = 0;
	for(int jj = 0; jj < JOY_MAX; jj++)
	{
		if(!joy_instance_array[jj].valid)
			continue;
		
		assert(pollfds_array[pollfd_next].fd == joy_instance_array[jj].fd);
		
		if(pollfds_array[pollfd_next].revents & (POLLHUP | POLLERR | POLLNVAL))
		{
			//This joystick's file descriptor is bad or encountered an error. Close it.
			printf("Lost joystick %d / player %d.\n", joy_instance_array[jj].num, joy_instance_array[jj].player);
			joy_instance_array[jj].valid = false;
			close(joy_instance_array[jj].fd);
			joy_instance_array[jj].fd = -1;
			joy_instance_array[jj].player = -1;
		}
		else if(pollfds_array[pollfd_next].revents & POLLIN)
		{
			//This joystick has an event to read. Read it.
			struct js_event event[16];
			ssize_t readret = read(joy_instance_array[jj].fd, event, sizeof(event));
			if(readret >= (ssize_t)sizeof(event[0]))
			{
				for(int ee = 0; ee < readret / (ssize_t)sizeof(event[0]); ee++)
				{
					//Handle the event, treating change/initial events identically.
					switch(event[ee].type & ~JS_EVENT_INIT)
					{
						case JS_EVENT_BUTTON:
							if(event[ee].number < BUTTON_MAX)
							{
								joy_instance_array[jj].button_state[event[ee].number] = event[ee].value;
							}
							break;
						case JS_EVENT_AXIS:
							if(event[ee].number < AXIS_MAX)
							{
								joy_instance_array[jj].axis_state[event[ee].number] = event[ee].value;
							}
							break;
						default:
							break;
					}
				}
			}
		}
		
		//File descriptors in pollfd array should be in the same order as valid joysticks were.
		pollfd_next++;
	}
}

//Updates JVS command buffer with latest data from joystick instances.
void command_update(void)
{	
	//Allow switching of mappings, if they have an activation chord
	for(int jj = 0; jj < JOY_MAX; jj++)
	{
		if(!joy_instance_array[jj].valid)
			continue;
		
		joy_instance_t *jptr = &(joy_instance_array[jj]);
		assert(jptr->joycfg >= 0 && jptr->joycfg < joycfg_count);
		joycfg_t *cptr = &(joycfg_array[jptr->joycfg]);
		
		//Iterate through all the mappings in the configured device, and look for active chords.
		assert(cptr->mapping_first >= 0);
		assert(cptr->mapping_first + cptr->mapping_count <= mapping_count);
		for(int mm = cptr->mapping_first; mm < cptr->mapping_first + cptr->mapping_count; mm++)
		{
			if(mapping_array[mm].activation_buttons_mask == 0)
				continue; //No chord defined
			
			if(mm == jptr->mapping)
				continue; //Already active
			
			uint32_t mask_needed = mapping_array[mm].activation_buttons_mask;
			for(int bb = 0; bb < BUTTON_MAX; bb++)
			{
				if(jptr->button_state[bb])
					mask_needed &= ~(1<<bb);
			}
			
			if(mask_needed == 0)
			{
				//All buttons in the chord were pressed, so activate the mapping
				printf("Activating mapping %d for joystick %d (player %d)\n", mm - cptr->mapping_first, jj, jptr->player);
				jptr->mapping = mm;
				break;
			}
		}
	}
	
	//Update activation of axis-as-button for each joystick, based on axis values
	for(int jj = 0; jj < JOY_MAX; jj++)
	{
		if(!joy_instance_array[jj].valid)
			continue;
		
		joy_instance_t *jptr = &(joy_instance_array[jj]);
		assert(jptr->mapping >= 0 && jptr->mapping < mapping_count);
		const mapping_t *mptr = &(mapping_array[jptr->mapping]);
		
		for(int aa = 0; aa < AXIS_MAX; aa++)
		{
			//Activate/deactivate in the negative direction, based on given thresholds			
			if(jptr->axis_state[aa] < mptr->daxis_neg_act[aa])
				jptr->daxis_neg_act[aa] = true;
			
			if(jptr->axis_state[aa] > mptr->daxis_neg_deact[aa])
				jptr->daxis_neg_act[aa] = false;
			
			//Activate/deactivate in the positive direction, based on given thresholds			
			if(jptr->axis_state[aa] > mptr->daxis_pos_act[aa])
				jptr->daxis_pos_act[aa] = true;
			
			if(jptr->axis_state[aa] < mptr->daxis_pos_deact[aa])
				jptr->daxis_pos_act[aa] = false;
		}
	}
	
	//Build the data we'll use for each player and for the system buttons...
	uint16_t userbuttons[4] = {0, 0, 0, 0};
	uint8_t sysbuttons = 0;
	
	//Run through each joystick and apply buttons and axes-as-buttons to the JVS switch data.
	for(int jj = 0; jj < JOY_MAX; jj++)
	{
		if(!joy_instance_array[jj].valid)
			continue;
		
		int mapping = joy_instance_array[jj].mapping;
		int player = joy_instance_array[jj].player;
		assert(mapping >= 0 && mapping < mapping_count);
		
		if(player < 0 || player >= 4)
			continue; //JVS-Bridge only supports 4 players
		
		
		for(int bb = 0; bb < BUTTON_MAX; bb++)
		{
			if(joy_instance_array[jj].button_state[bb])
			{
				userbuttons[joy_instance_array[jj].player] |= mapping_array[mapping].usrsw_for_button[bb];
				sysbuttons |= mapping_array[mapping].syssw_for_button[bb];
			}
		}
		
		for(int aa = 0; aa < AXIS_MAX; aa++)
		{
			if(joy_instance_array[jj].daxis_pos_act[aa])
				userbuttons[joy_instance_array[jj].player] |= mapping_array[mapping].daxis_pos_usrsw[aa];
			
			if(joy_instance_array[jj].daxis_neg_act[aa])
				userbuttons[joy_instance_array[jj].player] |= mapping_array[mapping].daxis_neg_usrsw[aa];
		}
		
	}
	
	//Build analog inputs, if any joystick has axes mapped.
	//Note - JVS expects unsigned 16-bit analog inputs. Linux gives us signed 16-bit inputs.
	uint16_t analogs[4] = {0, 0, 0, 0};
	
	
	//Assemble the command that we'll send to the JVS-Bridge, containing all the input data.
	//This format doesn't really have anything to do with JVS - it's just what the JVS-Bridge wants.
	//This could change in future firmware versions.
	commandbuf[0]  = 0xAA;
	commandbuf[1]  = 0x55;
	commandbuf[2]  = 0x01;
	commandbuf[3]  = sysbuttons;
	commandbuf[4]  = userbuttons[0] & 0xFF;
	commandbuf[5]  = (userbuttons[0] >> 8) & 0xFF;
	commandbuf[6]  = userbuttons[1] & 0xFF;
	commandbuf[7]  = (userbuttons[1] >> 8) & 0xFF;
	commandbuf[8]  = userbuttons[2] & 0xFF;
	commandbuf[9]  = (userbuttons[2] >> 8) & 0xFF;
	commandbuf[10] = userbuttons[3] & 0xFF;
	commandbuf[11] = (userbuttons[3] >> 8) & 0xFF;
	commandbuf[12] = analogs[0] / 256;
	commandbuf[13] = analogs[0] % 256;
	commandbuf[14] = analogs[1] / 256;
	commandbuf[15] = analogs[1] % 256;
	commandbuf[16] = analogs[2] / 256;
	commandbuf[17] = analogs[2] % 256;
	commandbuf[18] = analogs[3] / 256;
	commandbuf[19] = analogs[3] % 256;
	commandbuf[20] = 0;
	
	//I believe the NAOMI has a firmware bug dealing with 16-bit analog input values.
	//When the low byte is reported >= 0x80, the high byte seems to get decremented by the NAOMI.
	//I think someone accidentally used a signed addition when adding the low byte and high byte.
	//So, to prevent awful overflow bugs, just use 8-bit analog precision for now.
	commandbuf[13] = 0;
	commandbuf[15] = 0;
	commandbuf[17] = 0;
	commandbuf[19] = 0;
	
	//Compute the checksum
	commandbuf[21] = 0;
	for(int bb = 0; bb < 21; bb++)
	{
		commandbuf[21] += commandbuf[bb];
	}
	commandbuf[21] = -commandbuf[21];
}

//Entry
int main(int argc, const char **argv)
{
	printf("JVS-Bridge daemon %s\nRevision: %s\nBuilt by: %s\nDate: %s\n", RELEASEVERSION, BUILDVERSION, BUILDUSER, BUILDDATE);
	
	if(argc != 3)
	{
		printf("Usage: %s <Map File> <JVS-Bridge Device>\n", argv[0]);
		exit(-1);
	}
	
	//Read the mapfile
	printf("Reading mapfile.\n");
	mapfile_read(argv[1]);
		
	//Open the JVS-Bridge
	printf("Opening JVS-Bridge.\n");
	int bridge_fd = open(argv[2], O_RDWR | O_NOCTTY | O_SYNC);
	if(bridge_fd == -1)
	{
		perror("open");
		exit(-1);
	}
	
	//Read serial port settings to modify
	struct termios jvsattr;
	if(tcgetattr(bridge_fd, &jvsattr) == -1)
	{
		perror("tcgetattr");
		exit(-1);
	}
	
	//Enable raw mode
	cfmakeraw(&jvsattr);
	
	//Turn on hardware flow control
	jvsattr.c_iflag &= ~IXON;
	jvsattr.c_iflag &= ~IXOFF;
	jvsattr.c_cflag |= CRTSCTS;
	
	//Set to 8N1
	jvsattr.c_cflag &= ~CSIZE;
	jvsattr.c_cflag |= CS8;
	
	jvsattr.c_cflag &= ~PARENB;
	
	jvsattr.c_cflag &= ~CSTOPB;
	
	//Set to 115200 baud
	if(cfsetispeed(&jvsattr, B115200) == -1)
	{
		perror("cfsetispeed");
		exit(-1);
	}

	if(cfsetospeed(&jvsattr, B115200) == -1)
	{
		perror("cfsetospeed");
		exit(-1);
	}
	
	//Apply settings
	if(tcsetattr(bridge_fd, TCSANOW, &jvsattr) == -1)
	{
		perror("tcsetattr");
		exit(-1);
	}	
	
	//Main loop. Poll controls, build command, send to JVS-Bridge.
	int find_timer = 0;
	while(1)
	{
		//Try to find new joysticks periodically
		if(find_timer <= 0)
		{
			joy_find();
			find_timer = 100;
		}
		else
		{
			find_timer--;
		}
		
		//Handle joystick events
		joy_update();
		
		//Build control data
		command_update();
		
		//Update the JVS-Bridge with the new command
		write(bridge_fd, commandbuf, 22);
		tcflush(bridge_fd, TCIOFLUSH);
	}
}