//JVS-Bridge Public Release: Copyright 2022 Tucson Logic LLC.
//This program may be distributed under the terms of the GNU GPL v3.
//See the LICENSE file for more information.

//analog.c
//Analog input polling for JVS-Bridge analog variants
//Bryan E. Topp <bet@tucsonlogic.com> 2020

#include "analog.h"
#include "jvsbridge.h"
#include <SI_EFM8UB2_Register_Enums.h>

//Next analog channel to poll
static uint8_t analog_next = 0;

//Last values seem on analog channels
SI_SEGMENT_VARIABLE(analog_data[4], uint16_t, SI_SEG_XDATA);

void analog_init(void)
{
#if JVSBRIDGE_FWVAR_ANALOG

	//Switch the player-2 kick inputs (J6: p13, p14, p15, p16) to be analog inputs.
	//Analog input 0 - J6 Pin 13, EFM8UB2 pin 18, P4.4
	//Analog input 1 - J6 Pin 14, EFM8UB2 pin 17, P4.5
	//Analog input 2 - J6 Pin 15, EFM8UB2 pin 16, P4.6
	//Analog input 3 - J6 Pin 16, EFM8UB2 pin 15, P4.7
	P4MDIN  &= ~0xF0;
	P4      |=  0xF0;
	P4MDOUT &= ~0xF0;

	//Set ADC0 control...
	//Enable ADC (0x80), delayed-track mode (0x40).
	ADC0CN0 = 0xC0;

	//Set ADC0 config...
	//SAR clock frequency is limited to 8.33MHz, per the datasheet.
	//System clock is 48MHz, so we need to divide by 6. Set ADSC to 5.
	//Set to left-justified mode, so we get effectively 16-bit values from the ADC.
	ADC0CF = (5 << 3) | (1 << 2);

	//Set analog reference to VCC, 1x gain
	REF0CN = 0x88;

	//Set the analog negative input to ground
	AMX0N = 0x1F;

#endif //JVSBRIDGE_FWVAR_ANALOG
}

void analog_update(void)
{
#if JVSBRIDGE_FWVAR_ANALOG

	//See if we're still running a conversion on the ADC.
	//Don't update anything if so.
	if(ADC0CN0 & 0x10) //Check ADBUSY bit of ADC control register.
		return;

	//Read the value off the ADC
	//Do an exponential smoothing with the old data
	analog_data[analog_next] >>= 1;
	analog_data[analog_next] += ADC0H << 7;

	//We're done with that conversion now. Advance to the next one.
	analog_next = (analog_next + 1) % 4;

	//Set up the ADC input for the new conversion
	switch(analog_next)
	{
		case 0:
			//P4.4 as next input
			AMX0P = 13;
			break;
		case 1:
			//P4.5 as next input
			AMX0P = 14;
			break;
		case 2:
			//P4.6 as next input
			AMX0P = 15;
			break;
		case 3:
			//P4.7 as next input
			AMX0P = 34;
			break;
	}

	//Trigger the ADC to start the conversion by setting the ADBUSY bit.
	ADC0CN0 |= 0x10;

#endif //JVSBRIDGE_FWVAR_ANALOG
}

uint16_t analog_get(uint8_t ch)
{
#if JVSBRIDGE_FWVAR_ANALOG

	if(ch >= 4)
		return 0x8000;

	return analog_data[ch];

#else

	(void)ch;
	return 0x8000;

#endif
}
