//JVS-Bridge Public Release: Copyright 2022 Tucson Logic LLC.
//This program may be distributed under the terms of the GNU GPL v3.
//See the LICENSE file for more information.

//coinage.c
//Coin counting for JVS Bridge
//Bryan E. Topp <betopp@betopp.com> 2020

#include "coinage.h"
#include <SI_EFM8UB2_Register_Enums.h>

//Number of slots we support - should line up with the function reported in JVS ID
#define COINAGE_SLOTS 2

//Amount of coinage currently counted
uint16_t coinage_balances[COINAGE_SLOTS] = {0};

//Coin slot input debouncing
uint16_t coinage_slot_pressed_time[COINAGE_SLOTS] = {0};

//Number of main-loops we take to debounce coin slots.
//Todo - probably ought to synchronize this with a system timer or something.
#define COINAGE_DEBOUNCE_REQUIREMENT 10

void coinage_add(uint8_t slot, uint16_t amount)
{
	if(slot >= COINAGE_SLOTS)
		return;

	coinage_balances[slot] += amount;

	if(coinage_balances[slot] >= 16384) //Overflow detection - top 2 bits reserved per JVS spec
		coinage_balances[slot] = 16383;
}

void coinage_sub(uint8_t slot, uint16_t amount)
{
	if(slot >= COINAGE_SLOTS)
		return;

	if(coinage_balances[slot] < amount)
		coinage_balances[slot] = 0;
	else
		coinage_balances[slot] -= amount;
}

uint16_t coinage_get(uint8_t slot)
{
	if(slot >= COINAGE_SLOTS)
		return 0;

	return coinage_balances[slot];
}

void coinage_zero(uint8_t slot)
{
	if(slot >= COINAGE_SLOTS)
			return;

	coinage_balances[slot] = 0;
}

void coinage_update()
{
	//Coin slot 1 = Pin 39 = P1.7
	if(P1 & 0x80)
	{
		//Slot is currently pressed.
		if(coinage_slot_pressed_time[0] < COINAGE_DEBOUNCE_REQUIREMENT)
			coinage_slot_pressed_time[0]++;
	}
	else
	{
		//Slot has been released
		if(coinage_slot_pressed_time[0] >= COINAGE_DEBOUNCE_REQUIREMENT)
		{
			coinage_add(0, 1);
		}

		coinage_slot_pressed_time[0] = 0;
	}

	//Coin slot 2 = Pin 23 = P3.7
	if(P3 & 0x80)
	{
		//Slot is currently pressed.
		if(coinage_slot_pressed_time[1] < COINAGE_DEBOUNCE_REQUIREMENT)
			coinage_slot_pressed_time[1]++;
	}
	else
	{
		//Slot has been released
		if(coinage_slot_pressed_time[1] >= COINAGE_DEBOUNCE_REQUIREMENT)
		{
			coinage_add(1, 1);
		}

		coinage_slot_pressed_time[1] = 0;
	}
}
