//JVS-Bridge Public Release: Copyright 2022 Tucson Logic LLC.
//This program may be distributed under the terms of the GNU GPL v3.
//See the LICENSE file for more information.

//commands.c
//Processing of USB-UART input commands for JVS-Bridge
//Bryan E. Topp <betopp@betopp.com> 2020

#include <stdint.h>
#include <SI_EFM8UB2_Register_Enums.h>

#include "commands.h"
#include "commhw.h"
#include "jvsdefs.h"
#include "analog.h"
#include "jvsbridge.h"

#define COMMAND_BUFFER_LEN 32 //Length of a command packet that we'll receive from the USB host
SI_SEGMENT_VARIABLE(command_active[COMMAND_BUFFER_LEN], uint8_t, SI_SEG_XDATA); //Actual control data presented to JVS game
SI_SEGMENT_VARIABLE(command_buffer[COMMAND_BUFFER_LEN], uint8_t, SI_SEG_XDATA); //Data being received over USB at present
static uint8_t command_buffer_pos = 0; //How many bytes have already been received over USB into command_buffer
static uint8_t command_remain = 0; //How many further bytes we want to receive, before checking and applying command
static uint8_t command_sum = 0; //Running checksum of command received
static uint8_t command_start_step = 0; //Which of the command-start bytes we're expecting next
static uint8_t command_ever_ok = 0; //Whether we ever got a valid command over USB - if not, we'll use the switches.

//Used as a fallback if USB has never delivered a valid command - sets the active command based on switches.
static void commands_update_fromswitches()
{
	//System switches
	command_active[0] = 0;

	//J5 - Player 1 - EFM8 ports 1 and 2
	command_active[1] = 0;
	command_active[2] = 0;

	//J5.01 - UP - Pin 46 - P1.0
	if(!(P1 & 0x01)) { command_active[1] |= JVS_PLAYERSW0_UP; }
	//J5.02 - DN - Pin 45 - P1.1
	if(!(P1 & 0x02)) { command_active[1] |= JVS_PLAYERSW0_DOWN; }
	//J5.03 - RT - Pin 44 - P1.2
	if(!(P1 & 0x04)) { command_active[1] |= JVS_PLAYERSW0_RIGHT; }
	//J5.04 - LT - Pin 43 - P1.3
	if(!(P1 & 0x08)) { command_active[1] |= JVS_PLAYERSW0_LEFT; }
	//J5.05 - SL - Pin 42 - P1.4
	if(!(P1 & 0x10)) { command_active[1] |= JVS_PLAYERSW0_SVC; }
	//J5.06 - HM - Pin 41 - P1.5
	if(!(P1 & 0x20)) { command_active[0] |= 0x80; } //TEST
	//J5.07 - ST - Pin 40 - P1.6
	if(!(P1 & 0x40)) { command_active[1] |= JVS_PLAYERSW0_START; }
	//J5.08 - P1 - Pin 38 - P2.0
	if(!(P2 & 0x01)) { command_active[1] |= JVS_PLAYERSW0_PUSH1; }
	//J5.09 - P2 - Pin 37 - P2.1
	if(!(P2 & 0x02)) { command_active[1] |= JVS_PLAYERSW0_PUSH2; }
	//J5.10 - P3 - Pin 36 - P2.2
	if(!(P2 & 0x04)) { command_active[2] |= JVS_PLAYERSW1_PUSH3; }
	//J5.11 - P4 - Pin 35 - P2.3
	if(!(P2 & 0x08)) { command_active[2] |= JVS_PLAYERSW1_PUSH7; }
	//J5.13 - K1 - Pin 34 - P2.4
	if(!(P2 & 0x10)) { command_active[2] |= JVS_PLAYERSW1_PUSH4; }
	//J5.14 - K2 - Pin 33 - P2.5
	if(!(P2 & 0x20)) { command_active[2] |= JVS_PLAYERSW1_PUSH5; }
	//J5.15 - K3 - Pin 32 - P2.6
	if(!(P2 & 0x40)) { command_active[2] |= JVS_PLAYERSW1_PUSH6; }
	//J5.16 - K4 - Pin 31 - P2.7
	if(!(P2 & 0x80)) { command_active[2] |= JVS_PLAYERSW1_PUSH8; }

	//J6 - Player 2 - EFM8 ports 3 and 4
	command_active[3] = 0;
	command_active[4] = 0;

	//J6.01 - UP - Pin 30 - P3.0
	if(!(P3 & 0x01)) { command_active[3] |= JVS_PLAYERSW0_UP; }
	//J6.02 - DN - Pin 29 - P3.1
	if(!(P3 & 0x02)) { command_active[3] |= JVS_PLAYERSW0_DOWN; }
	//J6.03 - RT - Pin 28 - P3.2
	if(!(P3 & 0x04)) { command_active[3] |= JVS_PLAYERSW0_RIGHT; }
	//J6.04 - LT - Pin 27 - P3.3
	if(!(P3 & 0x08)) { command_active[3] |= JVS_PLAYERSW0_LEFT; }
	//J6.05 - SL - Pin 26 - P3.4
	if(!(P3 & 0x10)) { command_active[3] |= JVS_PLAYERSW0_SVC; }
	//J6.06 - HM - Pin 25 - P3.5
	if(!(P3 & 0x20)) { command_active[0] |= 0x80; } //TEST
	//J6.07 - ST - Pin 24 - P3.6
	if(!(P3 & 0x40)) { command_active[3] |= JVS_PLAYERSW0_START; }
	//J6.08 - P1 - Pin 22 - P4.0
	if(!(P4 & 0x01)) { command_active[3] |= JVS_PLAYERSW0_PUSH1; }
	//J6.09 - P2 - Pin 21 - P4.1
	if(!(P4 & 0x02)) { command_active[3] |= JVS_PLAYERSW0_PUSH2; }
	//J6.10 - P3 - Pin 20 - P4.2
	if(!(P4 & 0x04)) { command_active[4] |= JVS_PLAYERSW1_PUSH3; }
	//J6.11 - P4 - Pin 19 - P4.3
	if(!(P4 & 0x08)) { command_active[4] |= JVS_PLAYERSW1_PUSH7; }

#if !JVSBRIDGE_FWVAR_ANALOG //Analog variant uses these as analog inputs.
	//J6.13 - K1 - Pin 18 - P4.4
	if(!(P4 & 0x10)) { command_active[4] |= JVS_PLAYERSW1_PUSH4; }
	//J6.14 - K2 - Pin 17 - P4.5
	if(!(P4 & 0x20)) { command_active[4] |= JVS_PLAYERSW1_PUSH5; }
	//J6.15 - K3 - Pin 16 - P4.6
	if(!(P4 & 0x40)) { command_active[4] |= JVS_PLAYERSW1_PUSH6; }
	//J6.16 - K4 - Pin 15 - P4.7
	if(!(P4 & 0x80)) { command_active[4] |= JVS_PLAYERSW1_PUSH8; }
#endif // !JVSBRIDGE_FWVAR_ANALOG

}

void commands_init()
{
	uint8_t bb;
	for(bb = 0; bb < COMMAND_BUFFER_LEN; bb++)
	{
		command_active[bb] = 0;
		command_buffer[bb] = 0;
	}

	command_remain = 0;
	command_buffer_pos = 0;
	command_sum = 0;
}

void commands_update()
{
	//Check for a new byte available from the USB reception process.
    uint8_t newbyte = 0;
    if(!commhw_pcside_readb(&newbyte))
    {
    	//We don't have any new bytes over USB to process.

    	//If we've never gotten a valid command over USB, fall back to updating from the switch inputs.
    	if(!command_ever_ok)
    	{
    		commands_update_fromswitches();
    	}

    	return;
    }

    //Okay, we've got a new byte. What do we do with it?
	if(command_remain == 0)
	{
		//Not currently receiving a command. Check if it's time to receive a new one.
		//We need to receive a specific series of bytes to start buffering a command.
		if(command_start_step == 0)
		{
			//First byte to start a command - 0xAA
			if(newbyte == 0xAA)
				command_start_step = 1;
			else
				command_start_step = 0;
		}
		else if(command_start_step == 1)
		{
			//Second byte to start a command - 0x55
			if(newbyte == 0x55)
				command_start_step = 2;
			else
				command_start_step = 0;
		}
		else if(command_start_step == 2)
		{
			//Third byte to start a command - 0x01
			if(newbyte == 0x01)
			{
				command_remain = 19;
				command_buffer_pos = 0;
				command_sum = 0;
			}
			command_start_step = 0;
		}
		else
		{
			command_start_step = 0;
		}
	}
	else
	{
		//Receiving the command. Fill in the buffer and run the checksum.
		command_buffer[command_buffer_pos] = newbyte;
		command_sum += newbyte;
		command_buffer_pos++;
		command_remain--;

		//The final byte of the command, as received, is a parity byte that should zero our running checksum.
		//If that comes out correctly, apply the command.
		if(command_remain == 0 && command_sum == 0)
		{
			uint8_t bb;
			for(bb = 0; bb < COMMAND_BUFFER_LEN; bb++)
			{
				command_active[bb] = command_buffer[bb];
			}

			//We've got a valid update over USB - so stop polling the switch inputs from now on.
			command_ever_ok = 1;
		}
	}
}

uint8_t commands_getsyssw()
{
    return command_active[0];
}

uint8_t commands_getusrsw(uint8_t user, uint8_t bytenum)
{
    if(user >= 4)
        return 0;
    if(bytenum >= 2)
        return 0;

    return command_active[1 + (user*2) + bytenum];
}

uint16_t commands_getanalog(uint8_t channel)
{
	uint8_t bhigh;
	uint8_t blow;

	//If we're still reporting our own input data, give the analog value from ADC (if any)
	if(!command_ever_ok)
		return analog_get(channel) & 0xFF00; //NAOMI has strange firmware bug dealing with >8 bit precision

    if(channel >= 4)
        return 0x8000;
    
    bhigh = command_active[9 + (channel*2)];
    blow = command_active[10 + (channel*2)];
    return ((uint16_t)bhigh<<8u) + blow;
}

