//JVS-Bridge Public Release: Copyright 2022 Tucson Logic LLC.
//This program may be distributed under the terms of the GNU GPL v3.
//See the LICENSE file for more information.

//jvs.c
//JVS slave code for JVS Bridge
//Bryan E. Topp <betopp@betopp.com> 2020

#include "jvs.h"
#include "jvsdefs.h"
#include "commhw.h"
#include "commands.h"
#include "coinage.h"
#include "jvsbridge.h"
#include <SI_EFM8UB2_Register_Enums.h>

//Address assigned to this JVS slave
static uint8_t jvsslave_localaddr = 0; //0 is known invalid, as this identifies the master.

//Payload that we send for our slave ID response
static const char * const jvsslave_idstring = "Tucson Logic, LLC;JVS-Bridge;-Public-" JVSBRIDGE_FWDATE_s ";" JVSBRIDGE_FWVAR_s ;

//Payload that we send for our Function Check response
static const uint8_t jvsslave_fnstring[] = 
{
	0x01, //Switch inputs:
		4, //4 players
		16, //16 switches/player
		0,
	
	0x02, //Coin inputs:
		2, //2 coin slots
		0,
		0,
	
	0x03, //Analog inputs:
		4, //4 analog channels
		16, //16 effective bits
		0,	
	
	0x12, //GPIO:
		8, //8 fake outputs
		0,
		0,
	
	0x00 //End
};

//Whether we've received an escape character on the JVS bus, and the next received byte needs to be incremented.
static bool jvsslave_recv_escaped = false;

//Address received on JVS bus
static uint8_t jvsslave_recv_address = 0;
static bool jvsslave_recv_address_valid = false;

//Payload size received on JVS bus
static uint8_t jvsslave_recv_msgremain = 0;
static bool jvsslave_recv_msgremain_valid = false;

//Payload received on JVS bus - after separating sync/address and decoding escape characters.
#define JVSSLAVE_RECV_PAYLOAD_MAX 256
SI_SEGMENT_VARIABLE(jvsslave_recv_payload_array[JVSSLAVE_RECV_PAYLOAD_MAX], uint8_t, SI_SEG_XDATA);
static uint8_t jvsslave_recv_payload_count = 0;

//Checksum of packet received
static uint8_t jvsslave_recv_sum = 0;


//Response being constructed - after the sync, address, and length fields
#define JVSSLAVE_SEND_MAX 256
SI_SEGMENT_VARIABLE(jvsslave_send_array[JVSSLAVE_SEND_MAX], uint8_t, SI_SEG_XDATA);
static uint8_t jvsslave_send_count = 0;

//Sets the status code in the send buffer. Handles "prioritizing" worse responses.
static void jvsslave_send_buffer_status(jvs_status_t newstat)
{
	if(newstat > jvsslave_send_array[0])
		jvsslave_send_array[0] = newstat;
}

//Enqueues a byte in the send buffer if there's room
static void jvsslave_send_buffer_push(uint8_t value)
{
	if(jvsslave_send_count < JVSSLAVE_SEND_MAX)
	{
		jvsslave_send_array[jvsslave_send_count] = value;
		jvsslave_send_count++;
	}
	else
	{
		jvsslave_send_buffer_status(JVS_STATUS_ACKOVF);
	}
}

//Writes the given value on the JVS bus, escaping as necessary.
static void jvsslave_send_encoded(uint8_t val)
{
	if(val == JVS_BYTE_SYNC || val == JVS_BYTE_ESC)
	{
        commhw_jvsside_writeb(JVS_BYTE_ESC);
        commhw_jvsside_writeb(val-1);
	}
	else
	{
		commhw_jvsside_writeb(val);
	}
}

//Sends the send-buffer to the master, adding the address, size, and checksum
static void jvsslave_send_buffer_send()
{
	uint8_t dstaddr;
	uint8_t checksum;
	uint8_t ss;
	
	//Keep track of checksum as we write.
	//Checksum covers the destination address, payload+checksum size, and payload.
	//It happens before escape encoding.
	checksum = 0;
    commhw_jvsside_writeb(JVS_BYTE_SYNC);
	
	dstaddr = 0; //To master
	
	jvsslave_send_encoded(dstaddr);
	checksum += dstaddr;
	
	jvsslave_send_encoded(jvsslave_send_count+1);
	checksum += jvsslave_send_count+1;
	
	for(ss = 0; ss < jvsslave_send_count; ss++)
	{
		jvsslave_send_encoded(jvsslave_send_array[ss]);
		checksum += jvsslave_send_array[ss];
	}
	
	jvsslave_send_encoded(checksum);
}

//Processes a single JVS command. Returns how many bytes long the command was.
static int jvsslave_command(const uint8_t *cmdptr)
{
	//I despise SiLabs for shipping a compiler that doesn't support C99 variable declarations...
	uint8_t ii;
	uint8_t jj;
	const char *bb;

	switch(cmdptr[0])
	{
		case JVS_CMD_RESET:
			//Reset has a single mandatory parameter of 0xD9
			if(cmdptr[1] == 0xD9)
			{
				jvsslave_localaddr = 0;
                commhw_jvsside_senselow(false);
				jvsslave_send_buffer_status(JVS_STATUS_SILENT);
			}
			return 2;
		
		case JVS_CMD_ADDRSET:
			
            jvsslave_localaddr = cmdptr[1];
            commhw_jvsside_senselow(true);
            jvsslave_send_buffer_push(JVS_REPORT_NORMAL);
            
			return 2;

		case JVS_CMD_IOID:

			//Send textual ID
			jvsslave_send_buffer_push(JVS_REPORT_NORMAL);
			for(bb = jvsslave_idstring; *bb != '\0'; bb++)
			{
				jvsslave_send_buffer_push(*bb);
			}
			jvsslave_send_buffer_push(0);
			return 1;
			
		case JVS_CMD_CMDREV:
			jvsslave_send_buffer_push(JVS_REPORT_NORMAL);
			jvsslave_send_buffer_push(0x13); //Rev 1.3
			return 1;
			
		case JVS_CMD_JVSREV:
			jvsslave_send_buffer_push(JVS_REPORT_NORMAL);
			jvsslave_send_buffer_push(0x30); //Rev 3.0
			return 1;
		
		case JVS_CMD_COMREV:
			jvsslave_send_buffer_push(JVS_REPORT_NORMAL);
			jvsslave_send_buffer_push(0x10); //Rev 1.0
			return 1;
		
		case JVS_CMD_FNCHK:

			//Send function string
			jvsslave_send_buffer_push(JVS_REPORT_NORMAL);
			for(ii = 0; ii < sizeof(jvsslave_fnstring); ii++)
			{
				jvsslave_send_buffer_push(jvsslave_fnstring[ii]);
			}
			return 1;

		case JVS_CMD_INPUT_SW:
			jvsslave_send_buffer_push(JVS_REPORT_NORMAL);
			
			//Work through system buttons
			ii = commands_getsyssw();
			jvsslave_send_buffer_push(ii);
		
			//Work through all the players that were requested
			for(ii = 0; ii < cmdptr[1]; ii++)
			{
				//Work through all the switch bytes that were requested
				for(jj = 0; jj < cmdptr[2]; jj++)
				{
					uint8_t replybyte = commands_getusrsw(ii, jj);
					jvsslave_send_buffer_push(replybyte);
				}
			}
			return 3;
			
        case JVS_CMD_INPUT_ANALOG:
            jvsslave_send_buffer_push(JVS_REPORT_NORMAL);
            
            //Send all requested channels, 16 bits each
            for(ii = 0; ii < cmdptr[1]; ii++)
            {
                uint16_t val = commands_getanalog(ii);
                jvsslave_send_buffer_push(val / 256);
                jvsslave_send_buffer_push(val % 256);
            }
            return 2;
            
        case JVS_CMD_OUTPUT_GPDRV1:
            //Consume data and ignore driver output commands
            jvsslave_send_buffer_push(JVS_REPORT_NORMAL);
            return 2 + cmdptr[1];
        
        case JVS_CMD_OUTPUT_GPDRV2:
            //Ignore driver output
            jvsslave_send_buffer_push(JVS_REPORT_NORMAL);
            return 3;
            
		case JVS_CMD_OUTPUT_GPDRV3:
            //Ignore driver output
            jvsslave_send_buffer_push(JVS_REPORT_NORMAL);
            return 3;
		

		case JVS_CMD_PAYOUTBAL:
            //Return 0 for payout balance
            jvsslave_send_buffer_push(JVS_REPORT_NORMAL);
            jvsslave_send_buffer_push(0); //Hopper status
            jvsslave_send_buffer_push(0); //Hopper remaining (high)
            jvsslave_send_buffer_push(0); //Hopper remaining (mid)
            jvsslave_send_buffer_push(0); //Hopper remaining (low)
            return 2;
            
		

		case JVS_CMD_OUTPUT_COINSUB:
            coinage_sub(cmdptr[1], (((uint16_t)(cmdptr[2])) << 8) + cmdptr[3]);
            jvsslave_send_buffer_push(JVS_REPORT_NORMAL);
            return 4;
            
		case JVS_CMD_OUTPUT_COINADD:
			coinage_add(cmdptr[1], (((uint16_t)(cmdptr[2])) << 8) + cmdptr[3]);
			jvsslave_send_buffer_push(JVS_REPORT_NORMAL);
			return 4;

		case JVS_CMD_OUTPUT_PAYOUTADD:
            //Ignore payout addition
            jvsslave_send_buffer_push(JVS_REPORT_NORMAL);
            return 4;
         
        case JVS_CMD_INPUT_COIN:
        	//Return the requested number of coin counts
            jvsslave_send_buffer_push(JVS_REPORT_NORMAL);
            for(ii = 0; ii < cmdptr[1]; ii++)
            {
                jvsslave_send_buffer_push(coinage_get(ii) / 256);
                jvsslave_send_buffer_push(coinage_get(ii) % 256);
                coinage_zero(ii);
            }
            return 2;
            
		case JVS_CMD_INPUT_ROTARY:
		case JVS_CMD_INPUT_KEYCODE:
		case JVS_CMD_INPUT_SCRPOS:
		case JVS_CMD_INPUT_GPSW:
		case JVS_CMD_RETRY:
		case JVS_CMD_OUTPUT_ANALOG:
		case JVS_CMD_OUTPUT_CHAR:
		case JVS_CMD_OUTPUT_PAYOUTSUB:
		default:
			//Any unknown command necessarily stops processing - we don't know how long it is, to skip it.
			jvsslave_send_buffer_status(JVS_STATUS_BADCMD);
			return JVSSLAVE_RECV_PAYLOAD_MAX;
	}
}

//Checks the checksum of the payload in the payload array. Returns true if the checksum passed.
static bool jvsslave_checksum_valid()
{
	uint8_t received_checksum;
	uint8_t calculated_checksum;

	//The last byte of the payload is the checksum. So, we need at least one byte.
	if(jvsslave_recv_payload_count <= 0)
		return false;
	
	//We have accumulated the checksum byte as we received it.
	//So, subtracting it twice from the calculated sum should give 0.
	received_checksum = jvsslave_recv_payload_array[jvsslave_recv_payload_count-1];
	calculated_checksum = jvsslave_recv_sum;
	calculated_checksum -= received_checksum;
	calculated_checksum -= received_checksum;
	if(calculated_checksum == 0)
		return true;
	else
		return false;
}

//Reads the payload from the payload array and constructs replies to each message.
static void jvsslave_payload_process()
{
	uint8_t commands_length;
	uint8_t bytes_processed;
	uint8_t command_size;

	//Set up our response.
	//Initially, just contain an "okay" response code for the packet overall.
	jvsslave_send_array[0] = JVS_STATUS_NORMAL;
	jvsslave_send_count = 1;
	
	//Process all commands in the payload, or until one triggers an error.
	commands_length = jvsslave_recv_payload_count - 1;
	bytes_processed = 0;
	while(bytes_processed < commands_length)
	{
		command_size = jvsslave_command(&(jvsslave_recv_payload_array[bytes_processed]));
		//assert(command_size > 0);
		
		bytes_processed += command_size;
		
		if(jvsslave_send_array[0] == JVS_STATUS_BADCMD)
			break;
	}
	
	//Send a reply (only allowed if we have an address and didn't squelch it)
	if(jvsslave_localaddr != 0 && jvsslave_send_array[0] != JVS_STATUS_SILENT)
	{
        commhw_jvsside_txenable(true);
		jvsslave_send_buffer_send();
        commhw_jvsside_txenable(false);
	}
}

void jvsslave_init()
{
    commhw_jvsside_senselow(false);
    commhw_jvsside_txenable(false);
}

void jvsslave_update()
{
	//Read the next byte from the JVS bus.
    uint8_t newbyte = 0;
    if(!commhw_jvsside_readb(&newbyte))
        return;

	//If this is a sync word, reset the incoming buffer.
	if(newbyte == JVS_BYTE_SYNC)
	{
		jvsslave_recv_escaped = false;
		jvsslave_recv_address_valid = false;
		jvsslave_recv_msgremain_valid = false;
		jvsslave_recv_payload_count = 0;
		jvsslave_recv_sum = 0;
		return;
	}
	
	//If this is an escape character, note that, and await the next character
	if(newbyte == JVS_BYTE_ESC)
	{
		jvsslave_recv_escaped = true;
		return;
	}
	
	//This is not a sync or escape byte.
	//Apply the effects of the escape byte, if it was seen previously.
	if(jvsslave_recv_escaped)
	{
		newbyte++;
		jvsslave_recv_escaped = false;
	}
	
	//Keep a checksum of every data byte received since the last sync
	jvsslave_recv_sum += newbyte;
	
	//If we're waiting for an address, this is the address byte.
	if(jvsslave_recv_address_valid == false)
	{
		jvsslave_recv_address = newbyte;
		jvsslave_recv_address_valid = true;
		return;
	}
	
	//If we're waiting for the payload size, this is the payload size.
	if(jvsslave_recv_msgremain_valid == false)
	{
		jvsslave_recv_msgremain = newbyte;
		jvsslave_recv_msgremain_valid = true;
		return;
	}
	
	//Otherwise, we're receiving payload. Store it.
	if(jvsslave_recv_payload_count < JVSSLAVE_RECV_PAYLOAD_MAX)
	{
		jvsslave_recv_payload_array[jvsslave_recv_payload_count] = newbyte;
		jvsslave_recv_payload_count++;
	}
	
	//Count how many bytes of payload have been received, and process when the packet is finished.
	jvsslave_recv_msgremain--;
	if(jvsslave_recv_msgremain == 0)
	{
		//We've received all the bytes of the payload/checksum.
		//If the message was addressed to us and the checksum is valid, process it.
		if(jvsslave_recv_address == 0xFF || (jvsslave_localaddr != 0 && jvsslave_recv_address == jvsslave_localaddr))
		{
			if(jvsslave_checksum_valid())
			{
				jvsslave_payload_process();
			}
		}		
	}
}
