//JVS-Bridge Public Release: Copyright 2022 Tucson Logic LLC.
//This program may be distributed under the terms of the GNU GPL v3.
//See the LICENSE file for more information.

//jvsbridge-fwupd.cpp
//Firmware update utility for JVS-Bridge
//Copyright Tucson Logic, LLC 2020
//Bryan E. Topp <bet@tucsonlogic.com> 2020

//Uses wxWidgets for cross-platform GUI presentation.
#include <wx/wx.h>
#include <wx/aboutdlg.h>
#include <wx/filepicker.h>
#include <wx/ffile.h>
#include <wx/progdlg.h>

//Uses hidapi for communication with EFM8UB3 HID interface
#include <hidapi.h>

//Icon data
#include "tlllc.xpm"

//Vendor/product ID we look for, when looking for devices to update
#define FWUPD_VID 0x10C4 //SiLabs
#define FWUPD_PID 0xEACA //EFM8UB2 bootloader

class UpdaterApp : public wxApp
{
public:
	virtual bool OnInit();	
	virtual int OnExit();	

	//hidapi device
	hid_device *hidapi_handle;
};

class UpdaterFrame : public wxFrame
{
public:
	UpdaterFrame(const wxString &title, const wxPoint &pos, const wxSize &size);

private:
	//Callbacks from menu bar
	void OnExit(wxCommandEvent &event);
	void OnAbout(wxCommandEvent &event);

	//Callbacks from buttons
	void OnApply(wxCommandEvent &event);

	//File picker control that we use to choose the update file
	wxFilePickerCtrl *filePickerCtrl;

	wxDECLARE_EVENT_TABLE();
};

wxBEGIN_EVENT_TABLE(UpdaterFrame, wxFrame)
	EVT_MENU(wxID_EXIT, UpdaterFrame::OnExit)
	EVT_MENU(wxID_ABOUT, UpdaterFrame::OnAbout)
	EVT_BUTTON(wxID_APPLY, UpdaterFrame::OnApply)
wxEND_EVENT_TABLE()

wxIMPLEMENT_APP(UpdaterApp);

bool UpdaterApp::OnInit()
{
	//Initialize HIDAPI
	int hid_init_err = hid_init();
	if(hid_init_err != 0)
	{
		wxMessageBox( _("Failed to initialize HIDAPI."), _("HIDAPI failure"), wxICON_ERROR);
		return false;
	}
	
	//Try to find the device to update
	hidapi_handle = hid_open(FWUPD_VID, FWUPD_PID, NULL);
	if(hidapi_handle == NULL)
	{
		wxMessageBox( _(
			"Failed to find a JVS-Bridge device in update mode.\n"
			"Connect a JVS-Bridge while holding the \"FW UPD\" button.\n"
			"Then, run this program to perform the update."
		), _("Device not found"), wxICON_ERROR);
		return false;
	}
	
	//Then show the main window
	UpdaterFrame *frame = new UpdaterFrame(_("JVS-Bridge Firmware Updater"), wxDefaultPosition, wxDefaultSize);
	frame->Show(true);
	return true;
}

int UpdaterApp::OnExit()
{
	//Close the usb device
	if(hidapi_handle != nullptr)
	{
		hid_close(hidapi_handle);
		hidapi_handle = nullptr;
	}
	
	//Deinit HIDAPI
	hid_exit();
	
	//Success
	return 0;
}

UpdaterFrame::UpdaterFrame(const wxString &title, const wxPoint &pos, const wxSize &size)
	: wxFrame(NULL, wxID_ANY, title, pos, size)
{
	//Make icon
	SetIcon(wxICON(tlllc));
	(void)tlllc_xpm; //Suppress "unused" warning on Windows
	
	//Build menu bar
	wxMenu *menuFile = new wxMenu;
	menuFile->Append(wxID_EXIT);
	
	wxMenu *menuHelp = new wxMenu;
	menuHelp->Append(wxID_ABOUT);
	
	wxMenuBar *menuBar = new wxMenuBar;
	menuBar->Append(menuFile, _("&File"));
	menuBar->Append(menuHelp, _("&Help"));
	
	SetMenuBar(menuBar);

	wxPanel *mainPanel = new wxPanel(this);
	
	//Build window contents - input section
	wxStaticBoxSizer *inputSizer = new wxStaticBoxSizer(wxHORIZONTAL, mainPanel, _("Firmware File"));
	
	filePickerCtrl = new wxFilePickerCtrl(
		inputSizer->GetStaticBox(), 
		wxID_ANY, 
		wxEmptyString, 
		wxFileSelectorPromptStr,
		"EFM8UB2 Firmware (*.boot)|*.boot|All Files|*",
		wxDefaultPosition,
		wxDefaultSize,
		wxFLP_OPEN | wxFLP_FILE_MUST_EXIST | wxFLP_USE_TEXTCTRL);
		
	inputSizer->Add(filePickerCtrl, 1, wxEXPAND | wxALL, 4);

	//Build window contents - device section
	wxStaticBoxSizer *deviceSizer = new wxStaticBoxSizer(wxVERTICAL, mainPanel, _("JVS-Bridge Device"));

	wxStaticText *deviceText = new wxStaticText(
		deviceSizer->GetStaticBox(),
		wxID_ANY,
		_("Bootloader detected; firmware update ready."));
		
	deviceSizer->Add(deviceText, 1, wxEXPAND | wxALL, 4);

	wxButton *goButton = new wxButton(
		deviceSizer->GetStaticBox(),
		wxID_APPLY);
		
	deviceSizer->Add(goButton, 0, wxALL, 4);

	//Construct overall sizer
	wxBoxSizer *mainSizer = new wxBoxSizer(wxVERTICAL);
	mainSizer->Add(inputSizer, 0, wxEXPAND | wxALL, 4);
	mainSizer->Add(deviceSizer, 1, wxEXPAND | wxALL, 4);
	mainPanel->SetSizer(mainSizer);
	
	//Build status bar
	CreateStatusBar();
	SetStatusText(_("JVS-Bridge Firmware Updater ready."));
}

void UpdaterFrame::OnExit(wxCommandEvent &event)
{
	Close(true);
}

void UpdaterFrame::OnAbout(wxCommandEvent &event)
{
	wxAboutDialogInfo aboutInfo;
	aboutInfo.SetName(_("JVS-Bridge Firmware Updater"));
	aboutInfo.SetVersion(_("Version 1"));
	aboutInfo.SetDescription(_("Updates the firmware on a JVS-Bridge connected via USB."));
	aboutInfo.SetCopyright(_("(C) 2020 Tucson Logic, LLC"));
	aboutInfo.SetWebSite("http://tucsonlogic.com/");
	aboutInfo.AddDeveloper("Bryan E. Topp <bet@tucsonlogic.com>");
	wxAboutBox(aboutInfo);
}

void UpdaterFrame::OnApply(wxCommandEvent &event)
{
	//Check if they've actually selected a file.
	if(filePickerCtrl->GetPath().IsEmpty() || !filePickerCtrl->GetFileName().IsOk())
	{
		wxMessageBox(
			_("No firmware file selected. Please select a firmware file to write to the device."),
			_("No file selected"),
			wxICON_STOP);
		
		return;
	}
	
	//Try to open the file
	wxFFile FirmwareFile(filePickerCtrl->GetFileName().GetFullPath(), "rb");
	if(!FirmwareFile.IsOpened())
	{
		wxMessageBox(
			_("Failed to open the given firmware file. Ensure you have permission to read the file."),
			_("Failed to open file"),
			wxICON_STOP);
		
		return;
	}
	
	//Check the length and make sure it's sane
	size_t FWBytesLen = FirmwareFile.Length();
	if(FWBytesLen > 1048576)
	{
		wxMessageBox(
			_("The selected firmware file is over 1MB. This is not an EFM8UB2 firmware file."),
			_("Bad file size"),
			wxICON_STOP);
		return;
	}
	
	//Read in the contents of the file
	uint8_t *FWBytes = new uint8_t[FWBytesLen];
	if(FWBytes == nullptr)
	{
		wxMessageBox(
			_("Failed to allocate memory for reading the firmware file."),
			_("No memory"),
			wxICON_STOP);
		
		return;
	}
	
	size_t BytesRead = FirmwareFile.Read(FWBytes, FWBytesLen);
	if(BytesRead != FWBytesLen)
	{
		wxMessageBox(
			_("Failed to read the entire contents of the firmware file."),
			_("Failed to read file"),
			wxICON_STOP);
		
		return;
	}
	
	//Validate the commands contained in the file.
	//The format of the firmware file is described in SiLabs AN945.
	//In short, it is a series of records, each of which begin with a '$' character and a data-length byte.
	//Then, that many data bytes follow the length byte, and then the next record begins.
	//The record is sent to the bootloader on-chip, verbatim, so we don't particularly need to care about it.
	//We do some simple validation of the data, though, to make it less likely to brick the device.
	size_t FWPos = 0;
	bool ValidationFailed = false;
	size_t NumCommands = 0;
	while(FWPos != FWBytesLen) //We should end at exactly EOF
	{	
		if(FWPos + 2 >= FWBytesLen)
		{
			//Start of record runs off the end of the file
			//(Each record must be at least the start-of-record, size, and one data byte)
			ValidationFailed = true;
			break;
		}
		
		uint8_t StartByte = FWBytes[FWPos+0];
		uint8_t DataLen = FWBytes[FWPos+1];
		uint8_t CommandType = FWBytes[FWPos+2];
		
		//Check start-of-record byte
		if(StartByte != '$')
		{
			//Didn't find start-of-record where expected
			ValidationFailed = true;
			break;
		}
		
		//Check length byte
		if(DataLen < 2)
		{
			//Shouldn't have less than 2 data bytes
			ValidationFailed = true;
			break;
		}
		
		if(FWPos + DataLen + 2 > FWBytesLen)
		{
			//End of record runs off the end of the file
			ValidationFailed = true;
			break;
		}
		
		//The first of the data bytes is the command-type (interpreted by the bootloader).
		//Check the command-type - only allow specific commands.
		switch(CommandType)
		{
			case 0x31: //Setup
			case 0x32: //Erase
			case 0x33: //Write
			case 0x34: //Verify
			case 0x36: //RunApp
				break;
			
			//Others are not permitted - note, this includes the LOCK command.
			default:
				ValidationFailed = true;
				break;
		}
		
		//If the command-type is a write or erase, make sure it doesn't touch the bootloader sector
		if(CommandType == 0x32 || CommandType == 0x33)
		{
			if(DataLen < 3)
			{
				//Write or erase command must have at least the command-type and two address bytes
				ValidationFailed = true;
				break;
			}
			
			uint8_t addr_hi = FWBytes[FWPos+3];
			if(addr_hi >= 0xF5)
			{
				//Write/erase runs into the bootloader area of flash - disallow.
				ValidationFailed = true;
				break;
			}
		}
		
		//Skip to the next record - over the start-of-record and length, and all data.
		FWPos += 2 + FWBytes[FWPos+1];
		NumCommands++;
	}
	
	if(ValidationFailed)
	{
		wxMessageBox(
			_("The firmware file is corrupt or invalid. Please verify that the correct file is selected."),
			_("File contents bad"),
			wxICON_STOP);
		
		return;
	}
		
	//Actually run through the commands and send each to the device
	wxProgressDialog ProgDlg(_("Writing Firmware"), _("The selected firmware is being written to the device."), NumCommands, this);
	FWPos = 0;
	size_t CommandsCompleted = 0;
	hid_device *hidapi_handle = ((UpdaterApp*)wxTheApp)->hidapi_handle;
	while(FWPos < FWBytesLen)
	{
		wxASSERT(FWBytes[FWPos] == '$');
		uint8_t DataLen = FWBytes[FWPos+1] + 2;
		
		//Send the command as an HID SetReport.
		//Use multiple reports if we can't send all the data at once.
		uint8_t DataSent = 0;
		while(DataSent < DataLen)
		{
			uint8_t PacketSize = DataLen - DataSent;
			if(PacketSize > 64)
				PacketSize = 64;
			
			uint8_t packetbuf[65] = {0};
			packetbuf[0] = 0x00; //Report number
			memcpy(packetbuf+1, &(FWBytes[FWPos + DataSent]), PacketSize);
			
			printf("write %d:", PacketSize + 1);
			for(int pp = 0; pp < PacketSize + 1; pp++)
			{
				printf(" %2.2X", packetbuf[pp]);
			}
			printf("\n");
			
			int setreport_result = hid_write(hidapi_handle, packetbuf, PacketSize+1);
			if(setreport_result < 1)
			{
				wxMessageBox(
					_("Error sending payload to device."),
					_("USB Error"),
					wxICON_ERROR
				);
				
				return;
			}
			
			printf("hid_write returned %d\n", setreport_result);
			
			DataSent += setreport_result - 1;
		}
		
		printf("packet done\n");
				
		//Receive the response as an HID GetReport
		uint8_t ResponseBuffer[8] = {0};
		int getreport_result = hid_read(hidapi_handle, ResponseBuffer, 8);
		if(getreport_result <= 0)
		{
			wxMessageBox(
				_("Error receiving acknowledgement from device."),
				_("USB Error"),
				wxICON_ERROR
			);
			
			return;
		}
		
		//Check the response
		if(ResponseBuffer[0] != 0x40)
		{
			wxMessageBox(
				_("Device responded with an error code ") + wxString::Format("%d", ResponseBuffer[0]) + _(" after payload sent."),
				_("Device Error"),
				wxICON_ERROR
			);
			
			return;
		}
		
		//Move to the next command
		FWPos += DataLen;
		CommandsCompleted++;
		ProgDlg.Update(CommandsCompleted);
	}
	
	wxMessageBox(
		_("Firmware update completed."),
		_("Success")
	);
	
	Close(true);
}

