//JVS-Bridge Public Release: Copyright 2022 Tucson Logic LLC.
//This program may be distributed under the terms of the GNU GPL v3.
//See the LICENSE file for more information.

//jvsbridge-sdl2.c
//Example application for JVS Bridge - uses an SDL2 game controller.
//Bryan E. Topp <betopp@betopp.com> 2020

#include <SDL.h>
#include <unistd.h>
#include <fcntl.h>
#include <termios.h>
#include <sys/ioctl.h>
#include <errno.h>
#include <error.h>
#include <stdio.h>
#include <assert.h>


//Temporary buffer for making commands
#define COMMANDBUF_LEN 64
uint8_t commandbuf[COMMANDBUF_LEN];

//User switches that JVS supports, per seat.
typedef enum jvs_usrsw_e
{
	JVS_USRSW_PUSH2  =  0,
	JVS_USRSW_PUSH1  =  1,
	JVS_USRSW_RIGHT  =  2,
	JVS_USRSW_LEFT   =  3,
	JVS_USRSW_DOWN   =  4,
	JVS_USRSW_UP     =  5,
	JVS_USRSW_SVC    =  6,
	JVS_USRSW_START  =  7,
	
	JVS_USRSW_PUSH10 =  8,
	JVS_USRSW_PUSH9  =  9,
	JVS_USRSW_PUSH8  = 10,
	JVS_USRSW_PUSH7  = 11,
	JVS_USRSW_PUSH6  = 12,
	JVS_USRSW_PUSH5  = 13,
	JVS_USRSW_PUSH4  = 14,
	JVS_USRSW_PUSH3  = 15,
	
	JVS_USRSW_MAX
	
} jvs_usrsw_t;

//System switches that JVS supports, overall
typedef enum jvs_syssw_e
{
	JVS_SYSSW_RESERVED4 = 0,
	JVS_SYSSW_RESERVED3 = 1,
	JVS_SYSSW_RESERVED2 = 2,
	JVS_SYSSW_RESERVED1 = 3,
	JVS_SYSSW_TILT3     = 4,
	JVS_SYSSW_TILT2     = 5,
	JVS_SYSSW_TILT1     = 6,
	JVS_SYSSW_TEST      = 7,
	
	JVS_SYSSW_MAX
} jvs_syssw_t;

//Maps an SDL controller button into a JVS user-switch bitmask
jvs_usrsw_t usrsw_for_gcbtn(SDL_GameControllerButton btn)
{
	switch(btn)
	{
		case SDL_CONTROLLER_BUTTON_A:
			//return JVS_USRSW_PUSH1;
			return JVS_USRSW_DOWN;
			
		case SDL_CONTROLLER_BUTTON_B:
			//return JVS_USRSW_PUSH2;
			return JVS_USRSW_UP;
			
		case SDL_CONTROLLER_BUTTON_X:
			return JVS_USRSW_PUSH3;
			
		case SDL_CONTROLLER_BUTTON_Y:
			return JVS_USRSW_PUSH4;
			
		case SDL_CONTROLLER_BUTTON_START:
			return JVS_USRSW_START;
		
		case SDL_CONTROLLER_BUTTON_BACK:
			return JVS_USRSW_SVC;
			
		case SDL_CONTROLLER_BUTTON_LEFTSHOULDER:
			return JVS_USRSW_PUSH5;
			
		case SDL_CONTROLLER_BUTTON_RIGHTSHOULDER:
			return JVS_USRSW_PUSH6;
			
		case SDL_CONTROLLER_BUTTON_DPAD_UP:
			return JVS_USRSW_UP;
		
		case SDL_CONTROLLER_BUTTON_DPAD_DOWN:
			return JVS_USRSW_DOWN;
		
		case SDL_CONTROLLER_BUTTON_DPAD_LEFT:
			return JVS_USRSW_LEFT;
		
		case SDL_CONTROLLER_BUTTON_DPAD_RIGHT:	
			return JVS_USRSW_RIGHT;
		
		default:
			return JVS_USRSW_MAX;
	}
}

//Maps an SDL controller button into a JVS system-switch bitmask
jvs_syssw_t syssw_for_gcbtn(SDL_GameControllerButton btn)
{
	switch(btn)
	{			
		case SDL_CONTROLLER_BUTTON_GUIDE:
			return JVS_SYSSW_TEST;

		default:
			return JVS_SYSSW_MAX;
	}
}

//Maps an SDL controller axis to a JVS axis number
int analog_for_gcaxis(SDL_GameControllerAxis axis)
{
	switch(axis)
	{
		case SDL_CONTROLLER_AXIS_LEFTX:
			return 0;
		
		case SDL_CONTROLLER_AXIS_TRIGGERLEFT:
			return 2;
		
		case SDL_CONTROLLER_AXIS_TRIGGERRIGHT:
			return 1;
			
		default:
			return -1;
	}

}

//Game-controller used by SDL
static SDL_GameController *gcptr = NULL;

//Current state of player switches
static uint16_t usrmask = 0;

//Current state of system switches
static uint8_t sysmask = 0;

//Current analog values
static uint16_t analogs[4] = {0};


int main(int argc, const char **argv)
{
	if(argc != 2)
	{
		printf("usage: %s <device>\n", argv[0]);
		return 0;
	}
	
	//Need to make a dummy video output for SDL
	if(SDL_InitSubSystem(SDL_INIT_VIDEO) != 0)
	{
		fprintf(stderr, "Failed to init SDL video\n");
		exit(-1);
	}
	
	SDL_Window *window;
	SDL_Renderer *renderer;
	if(SDL_CreateWindowAndRenderer(100, 100, SDL_WINDOW_RESIZABLE, &window, &renderer) != 0)
	{
		fprintf(stderr, "Failed to make SDL window/renderer\n");
		exit(-1);
	}
	
	int bridge_fd = open(argv[1], O_RDWR | O_NOCTTY | O_SYNC);
	if(bridge_fd == -1)
	{
		fprintf(stderr, "Error opening device %s\n", argv[1]);
		exit(-1);
	}
	
	
	//Read serial port settings to modify
	struct termios jvsattr;
	if(tcgetattr(bridge_fd, &jvsattr) == -1)
	{
		perror("tcgetattr");
		exit(-1);
	}
	
	//Enable raw mode
	cfmakeraw(&jvsattr);
	
	//Turn on hardware flow control
	jvsattr.c_iflag &= ~IXON;
	jvsattr.c_iflag &= ~IXOFF;
	jvsattr.c_cflag |= CRTSCTS;
	
	//Set to 8N1
	jvsattr.c_cflag &= ~CSIZE;
	jvsattr.c_cflag |= CS8;
	
	jvsattr.c_cflag &= ~PARENB;
	
	jvsattr.c_cflag &= ~CSTOPB;
	
	//Set to 115200 baud
	if(cfsetispeed(&jvsattr, B115200) == -1)
	{
		perror("cfsetispeed");
		exit(-1);
	}

	if(cfsetospeed(&jvsattr, B115200) == -1)
	{
		perror("cfsetospeed");
		exit(-1);
	}
	
	//Apply settings
	if(tcsetattr(bridge_fd, TCSANOW, &jvsattr) == -1)
	{
		perror("tcsetattr");
		exit(-1);
	}
	
	
	
	
	//Open the gamepad and start polling it for updates
	if(SDL_InitSubSystem(SDL_INIT_GAMECONTROLLER) != 0)
	{
		fprintf(stderr, "Failed to init SDL game controller system.\n");
		exit(-1);
	}
	
	gcptr = SDL_GameControllerOpen(0);
	if(gcptr == NULL)
	{
		fprintf(stderr, "Failed to open first SDL game controller.\n");
		exit(-1);
	}
	
	while(1)
	{
		SDL_Event e;
		while(SDL_PollEvent(&e))
		{
		
			switch(e.type)
			{
				case SDL_QUIT:
					exit(0);
					break;
				
				case SDL_CONTROLLERAXISMOTION:
				{
					int axisnum = analog_for_gcaxis(e.caxis.axis);
					if(axisnum >= 0 && axisnum < 4)
					{
						//if(axisnum == 0)
						//	printf("%d\n", e.caxis.value);
					
						analogs[axisnum] = e.caxis.value + 32768;
					}
					
					break;
				}
					
				case SDL_CONTROLLERBUTTONDOWN:				
				case SDL_CONTROLLERBUTTONUP:
				{
					jvs_syssw_t syssw = syssw_for_gcbtn(e.cbutton.button);
					if(syssw < JVS_SYSSW_MAX)
					{
						if(e.cbutton.state == SDL_PRESSED)
							sysmask |= (1<<syssw);
						else
							sysmask &= ~(1<<syssw);
						
						
					}
					
					jvs_usrsw_t usrsw = usrsw_for_gcbtn(e.cbutton.button);
					if(usrsw < JVS_USRSW_MAX)
					{
						if(e.cbutton.state == SDL_PRESSED)
							usrmask |= (1<<usrsw);
						else
							usrmask &= ~(1<<usrsw);
					}
					
					break;
				}
				default:
					break;
				
			}
		}
		
		//Assemble the command that we'll send to the JVS-Bridge, containing all the input data.
		commandbuf[0] = 0xAA;
		commandbuf[1] = 0x55;
		commandbuf[2] = 0x01;
		commandbuf[3] = sysmask;
		commandbuf[4] = usrmask & 0xFF;
		commandbuf[5] = (usrmask >> 8) & 0xFF;
		commandbuf[6] = 0; //p2 first byte
		commandbuf[7] = 0; //p2 second byte
		commandbuf[8] = 0; //p3 first byte
		commandbuf[9] = 0; //p3 second byte
		commandbuf[10] = 0; //p4 first byte
		commandbuf[11] = 0; //p4 second byte
		commandbuf[12] = analogs[0] / 256;
		commandbuf[13] = analogs[0] % 256;
		commandbuf[14] = analogs[1] / 256;
		commandbuf[15] = analogs[1] % 256;
		commandbuf[16] = analogs[2] / 256;
		commandbuf[17] = analogs[2] % 256;
		commandbuf[18] = analogs[3] / 256;
		commandbuf[19] = analogs[3] % 256;
		commandbuf[20] = 0;
		
		//I believe the NAOMI has a firmware bug dealing with 16-bit analog input values.
		//When the low byte is reported >= 0x80, the high byte seems to get decremented by the NAOMI.
		//I think someone accidentally used a signed addition when adding the low byte and high byte.
		//So, to prevent awful overflow bugs, just use 8-bit analog precision for now.
		commandbuf[13] = 0;
		commandbuf[15] = 0;
		commandbuf[17] = 0;
		commandbuf[19] = 0;
		
		//Compute checksum
		commandbuf[21] = 0;
		for(int bb = 0; bb < 21; bb++)
		{
			commandbuf[21] += commandbuf[bb];
		}
		commandbuf[21] = -commandbuf[21];
		
		//Update the JVS-Bridge with the new command
		write(bridge_fd, commandbuf, 22);
		tcflush(bridge_fd, TCIOFLUSH);
		
		//Hold off on updates momentarily
		SDL_Delay(10);
		SDL_RenderClear(renderer);
		SDL_RenderPresent(renderer);
	}
	
	
	return 0;
}
